##################
#
# klasa bibSearch
#
# klasa suy do rekursywnego obliczania wielkoci katalogu wraz z podkatalogami
#

package bibSearch;

use strict;
use dirTree;
use utilities;
use vars qw(@ISA);


@ISA = ("dirTree");

#############################################

sub trimExternalPunctation{
  my ($entry)=@_;
  while($entry =~ /^\s*[.,]\s*(.*)$/){
    $entry=$1;
  }
  while($entry =~ /^(.*)\s*[,]\s*$/){
    $entry=$1;
  }
  return $entry;
}

#############################################

sub trimExternalBraces{
  my ($entry)=@_;
  while($entry =~ /^\s*\{\s*(.*)$/){
    $entry=$1;
  }
  while($entry =~ /^(.*)\s*\}\s*$/){
    $entry=$1;
  }
  return $entry;
}

#############################################

sub getNameKey{
  my ($nameKey)=@_;
  $nameKey =~ s/[.{}\\\'\"]//g;
  return $nameKey;
}

#############################################

sub getTitleKey{
  my ($titleKey)=@_;
  $titleKey =~ s/[-:()*&_+\/?<>.,{}\\\'\"]//g;  #"
  $titleKey =~ s/\s+/-/g;
  $titleKey = lc($titleKey);
  return $titleKey;
}

#############################################

sub separateLabel{
  my ($entry)=@_;
  if($entry =~ /^([^}]+)\}(.*)$/){
    return ($1,$2);
  }else{
    return ("",$entry);
  }
}

#############################################

sub separateName{
  my ($entry)=@_;
  # remove the leading \\
  while($entry =~ /^(\\ |\s+|,|and\s+|\\&\s+)(.*)$/){
    $entry=$2;
  }
  # remove the leading initials
  my $initials='';
  while($entry =~ /^([A-Z]\.)\\?\s*(.*)$/){
    $initials.=$1;
    $entry=$2;
  }
  my $surname='';
  if($initials && $entry =~ /^([.A-Za-z\\'\{\}"]+)(.*)$/){  #"
    return ($initials,$1,$2);
  }else{
    return ('','',$entry);
  }
}

#############################################

sub separatePreprint{
  my ($entry)=@_;
  if($entry =~ /^(.*)(,| |, )\s*(in\s+preparation|preprint|manuscript|work\s+in\s+progress)\s*$/){
    return ($3,$1);
  }else{
    return ('',$entry);
  }
}

#############################################

sub separateAccepted{
  my ($entry)=@_;
  if($entry =~ /^(.*)(,| |, )\s*to\s+appear\s+in\s+(.*)\s*$/){
    return ('to appear',"$1, $3");
  }elsif($entry =~ /^(.*)(,| |, )\s*(to\s+appear|accepted|submitted|in\s+print)\s*$/){
    return ($3,$1);
  }else{
    return ('',$entry);
  }
}

#############################################

sub separatePages{
  my ($entry)=@_;
  if($entry =~ /^(.*)(,| |, )\s*(\d+)-+(\d+)(|[., ].*)\s*$/){
    return ("$3--$4","$1 $5");
  }else{
    return ('',$entry);
  }
}

#############################################

sub separateBookYear{
  my ($entry)=@_;
  if($entry =~ /^(.*)\s*([12][7890]\d\d)$/){
    return ($2,$1);
  }else{
    return ('',$entry);
  }
}

#############################################

sub separateJournalYear{
  my ($entry)=@_;
  if($entry =~ /^(.*)\s*\(([12][7890]\d\d)\)\s*(.*)$/){
    return ($2,"$1 $3");
  }else{
    return ('',$entry);
  }
}

#############################################

sub separateBookNumber{
  my ($entry)=@_;
  if($entry =~ /^(.*)\s*\((\d\d?)\)\s*(.*)$/){
    return ($2,"$1 $3");
  }elsif($entry =~ /^(.*)\s*[nN]o\.\s*(\d\d?),?\s*(.*)$/){
    return ($2,"$1 $3");
  }else{
    return ('',$entry);
  }
}

#############################################

sub separateVolumeAndYear{
  my ($entry)=@_;
  if($entry =~ /^(.*)\s*\{\\bf\s+(\d+[A-Z*]?)\},?\s*\(([12][7890]\d\d)\)\s*(.*)$/){
    return ($2,$3,"$1 $4");
  }elsif($entry =~ /^(.*)\s+(\d\d?\d?[A-Z*]?)\s*\(([12][7890]\d\d)\)\s*(.*)$/){
    return ($2,$3,"$1 $4");
  }else{
    return ('','',$entry);
  }
}

#############################################

sub separateVolume{
  my ($entry)=@_;
  if($entry =~ /^(.*)\s*\{\\bf\s+(\d+[A-Z*]?)\},?\s*$/){
    return ($2,"$1 $3");
  }elsif($entry =~ /^(.*)Vol\.\s+(\d\d?\d?[A-Z*]?)\s*(.*)$/){
    return ($2,"$1 $3");
  }else{
    return ('','',$entry);
  }
}

#############################################

sub separateJournalName{
  my ($entry)=@_;
  $entry=trimExternalPunctation($entry);
  # {\em xxx} or {\it xxx}  case
  if($entry =~ /^(.*)\s*\{\s*\\(it|em)\\?\s+([^}]+)\s*\}\s*$/){
    return ($3,$1);
  # , J. Journal etc.  case
  }elsif($entry =~ /^(.*)\s*,\s*(J\.|Journal|Proceedings|Transactions)(.+)\s*$/){
    return ("$2$3",$1);
  # , xxx  case
  }elsif($entry =~ /^(.*)\s*,\s*([^,]+)\s*$/){
    return ($2,$1);
  # ,} xxx  case
  }elsif($entry =~ /^(.*)\s*,\s*\}\s*([^}]+)\s*$/){
    return ($2,"$1}");
  }else{
    return ('',$entry);
  }
}

#############################################

sub separateTitle{
  my ($entry)=@_;
  $entry=trimExternalPunctation($entry);
  my $title;
  # {\em xxx} or {\it xxx}  case
  if($entry =~ /^\s*\{\s*\\(it|em)\s+([^}]+)\s*\}\s*(.*)$/){
    $title=$2;
    $entry=$3;
  # xxx,  case
  }elsif($entry =~ /^\s*([^,]+)\s*,\s*(.*)$/){
    $title=$1;
    $entry=$2;
  # xxx  case
  }elsif($entry =~ /^\s*(\S.*\S)\s*$/){
    $title=$1;
    $entry='';
  # empty  case
  }else{
    $title='';
    $entry='';
  }
  $title =~ s/\s+/ /g;
  $title=trimExternalBraces($title);
  $title=trimExternalPunctation($title);
  return ($title,$entry);
}

#############################################

sub extractTitle{
  my ($entry)=@_;
  $entry=trimExternalPunctation($entry);
  my $title;
  # {\em xxx} or {\it xxx}  case
  if($entry =~ /^\s*\{\\(it|em)\s+([^}]+)\s*\}\s*$/){
    $title=$2;
  }else{
    $title=$entry;
  }
  $title =~ s/\s+/ /g;
  $title=trimExternalBraces($title);
  $title=trimExternalPunctation($title);
  return ($title,'');
}

#############################################
#
# analyze
#
sub analyze{
  my ($self,$entry,$prefix,$file)=@_;
  my $bibRec={};
  my $origEntry="$prefix$entry";
  while($origEntry =~ /^(.*)\s*[\n\015]+\s*$/){
    $origEntry=$1;
  }
  while($entry =~ /^(.*)%[^\015]*\015(.*)$/){
    $entry="$1 $2";
  }
  $entry =~ s/\015/ /g;
  while($entry =~ /^(.*)%.*$/){
    $entry=$1;
  }
  $bibRec->{entry}=$entry;
  # Process label
  my $label;
  ($label,$entry)=separateLabel($entry);

  # Process names
  my ($initials,$surname);
  my $authors='';
  my $namesKey='';
  while(1){
    ($initials,$surname,$entry)=separateName($entry);
    last if(!$initials && !$surname);
    $namesKey.=getNameKey($surname);
    $authors.="$initials\\ $surname, "
  }
  if($authors =~ /^(.*)\s*,\s*$/){
    $authors=$1;
  }

  # Process final fullstop
  if($entry =~ /^(.*)\s*\.\s*$/){
    $entry=$1;
  }

  my $type="unrecognized";
  my $collection='';
  my $edited='';
  if($entry =~ /\d+--?\d+/){
    $collection='yes';
  }
  if($entry =~ /\([12][7890]\d\d\)/){
    $collection='yes';
    $type="journal";
  }elsif($entry =~ /[A-Z]\w+\s+[12][7890]\d\d/){
    $type='book';
    $collection='no';
  }elsif($entry =~ /[12][7890]\d\d/){
    $collection='no';
  }
  if($entry =~ /Ph.D.\s*Thesis/){
    $collection='no';
    $type="PhD";
  }
  if($entry =~ /Editor|ed./){
    $collection='yes';
    $edited=$1;
  }

  my $pages;
  my $bookNumber;
  my $journalYear;
  my $volume;
  my $journalName;
  my $title;
  my $descriptor;
  my $bookYear;
  my $bookInfo;
  my $level=2; # 2 - published, 1 - accepted, 0 - in preparation

  if(!$type){
    # Process preprint article
    ($descriptor,$entry)=separatePreprint($entry);
    if($descriptor){
      ($title,$entry)=extractTitle($entry);
      $type="preprintArticle";
      $level=0;
    }

    # Process accepted article
    ($descriptor,$entry)=separateAccepted($entry);
    if($descriptor){
      ($journalName,$entry)=separateJournalName($entry);
      ($title,$entry)=extractTitle($entry);
      $type="acceptedArticle";
      $level=1;
    }
  }

  # Process journal article
    if($type eq "journal"){
    ($pages,$entry)=separatePages($entry);
    if($pages){
      ($bookNumber,$entry)=separateBookNumber($entry);
      ($volume,$journalYear,$entry)=separateVolumeAndYear($entry);
      if(!$volume){
        ($journalYear,$entry)=separateJournalYear($entry);
        ($volume,$entry)=separateVolume($entry);
      }
      ($journalName,$entry)=separateJournalName($entry);
      ($title,$entry)=extractTitle($entry);
    }
    $type="journalArticle";
  }

  # Process book
  if($type !~ "Article"){
    ($bookYear,$entry)=separateBookYear($entry);
    ($title,$entry)=separateTitle($entry);
    $type="book";
    $bookInfo=trimExternalPunctation($entry);
    $bookInfo =~ s/\s+/ /g;
  }

  return if($type eq "unrecognized" || $namesKey !~ /\S/);

  my $year=($journalYear ? $journalYear : $bookYear);
  my $titleKey=getTitleKey($title);
  my $key="$namesKey:$year:$titleKey";

  my $signature="$key:$pages";
  $bibRec->{type}=$type;
  $bibRec->{authors}=$authors;
  $bibRec->{title}=$title;
  $bibRec->{journalName}=$journalName;
  $bibRec->{year}=$year;
  $bibRec->{volume}=$volume;
  $bibRec->{bookNumber}=$bookNumber;
  $bibRec->{pages}=$pages;
  $bibRec->{level}=$level;
  $bibRec->{bookInfo}=$bookInfo;
  $bibRec->{origEntry}=$origEntry;
  $bibRec->{source}=$file;
  $bibRec->{bibKey}="$namesKey-$bibRec->{year}";

  if(!defined($main::items->{$key})){
    $main::items->{$key}=$bibRec;
  }else{
    if($bibRec->{level} > $main::items->{$key}->{level}){
      $main::items->{$key}=$bibRec;
    }elsif($bibRec->{signature} ne $main::items->{$key}->{signature}){
      print "New record with inconsistent signature:\n";
      print "   Existing signature: $main::items->{$key}->{signature}\n";
      print "   New signature: $bibRec->{signature}\n";
    }
  }


  if($namesKey){
  }else{
    print "/*/*/*> unrecognized\n$entry\n\n";
  }
}

#############################################
#
# processFile
#
sub processFile{
  my ($self,$fullFileName)=@_;
  if($fullFileName !~ /\.tex$/){
    return 0;
  }
  open(IN,"<$fullFileName") || die("bibSearch: Failed to open file $fullFileName\n");

  my $size = -s $fullFileName;
  my $text;

  binmode(IN);
  read(IN,$text,$size);
  $text =~ s/\032/ /g;
  $text =~ s/\012/ /g;

  if($text =~ /\\begin\{thebibliography\}(.*)\\end\{thebibliography\}/){
    return 0 if($text =~ /% no bibSearch/);
    $text=$1;
    my $prefix="\\bibitem\{";
    my @entries=split(/\\bibitem\{/,$text);
    foreach my $entry (@entries){
      $self->analyze($entry,$prefix,$fullFileName);
    }
    my $n=$#entries+1;
    print "$n entries found in $fullFileName \n";
    return 1;
  }
  return 0;
}

1;